import importlib.machinery
import importlib.resources as pkg_resources
import importlib.util
from datetime import datetime
from pathlib import Path
from typing import Type

import click

from aoc import assets
from aoc.solution import Solution
from aoc.solution import SolutionResult


def read_asset(asset_name: str) -> str:
    """Read an asset file"""
    return pkg_resources.read_text(assets, asset_name)


def get_latest_year() -> int:
    """Returns the last year if we are not in december yet otherwise returns the current year."""
    today = datetime.now()
    if today.month < 12:
        return today.year - 1
    return today.year


def get_solution_class(solution_file: Path) -> Type[Solution]:
    """Import the correct solution"""

    if not solution_file.exists():
        raise FileNotFoundError(f"Solution file {solution_file} not found!")

    module_name = f"solutions_{solution_file.name[:-3]}"
    cls_name = solution_file.name[:-3].title()

    spec = importlib.util.spec_from_file_location(module_name, solution_file.resolve())

    if not spec:
        raise ImportError(f"Could not import {solution_file}")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)  # type: ignore

    return getattr(module, cls_name)


def get_input_data(input_file: Path) -> list[str]:
    """Get the input data from the file"""
    if not input_file.exists():
        raise FileNotFoundError(f"Input file {input_file} not found!")

    with open(input_file, "r") as f:
        return f.readlines()


def instantiate_solution(solution_file: Path, input_file: Path) -> Solution:
    """Import the correct solution and return an instance of it"""

    solution_cls = get_solution_class(solution_file)
    input_data = get_input_data(input_file)

    solution = solution_cls()  # type: ignore
    solution.set_input_data(input_data)
    return solution


def display_solution(solution: Solution, timeit: bool, number: int) -> None:
    """Display the first and second parts of the solution"""
    part_one, part_two = solution(timeit, number)

    def part_name(part: int) -> str:
        color = "green" if part == 1 else "red"
        return click.style(f"Part {part}:", fg=color)

    def display_part(part: int, result: SolutionResult) -> None:
        time_res = f" ({result.avg_time:.05f} sec)" if timeit else ""
        click.echo(f"{part_name(part)} {result.result}{time_res}")

    click.secho(solution, bold=True)

    display_part(1, part_one)
    display_part(2, part_two)

    if timeit:
        click.secho(f"Times represent average of {number:,d} runs", fg="yellow")



"""The abstract solution class"""
from abc import ABC
from abc import abstractmethod
from timeit import timeit as get_best_time
from typing import Any
from typing import Optional

from aoc.input_handler import InputHandler


class SolutionResult:
    """Class for storing the result of a solution."""

    def __init__(
        self, result: Any, time: Optional[float], number: Optional[int]
    ) -> None:
        """Initialize the solution result.

        Parameters
        ----------
        result : Any
            The result of the solution.

        time : Optional[float]
            The time taken to solve the solution.

        number : Optional[int]
            The number of repetitions when calculating the time.
        """
        self.result = result
        self.time = time
        self.number = number

    @property
    def avg_time(self) -> float:
        """Return the average time of execution. Returns -1 if not timing."""
        if not self.time:
            return -1

        time = self.time if self.time else 0
        number = self.number if self.number else 0
        return time / number

    def __str__(self) -> str:
        """Return the string representation of the solution result."""
        return f"{self.result}"

    def __repr__(self) -> str:
        return self.__str__()


class Solution(ABC):
    """Base class for a solution

    This class should be inherited by concrete solutions.
    There are varying functions that must be, and can be implemented.

    You can call the class to create a tuple of results from part one and part two.

    Parameters
    ----------
    year : int
        The year of the problem

    day : int
        The day of the problem

    name : Optional[str], optional
        The fun name of the problem, by default None

    """

    def __init__(self, year: int, day: int, name: str):
        self.year = year
        self.day = day
        self.name = name

        self.input: InputHandler
        self.__data: list[Any]

    @property
    def data(self) -> Any:
        """The data, reformated and mutated"""

        if not self.__data and isinstance(self.__data, list):
            raise ValueError("Data should not be empty")

        return self.__data

    def set_input_data(self, input_data: list[str]) -> None:
        """Reload and reformat the input file.

        Parameters
        ----------
        input_data : list[str]
            The input data as a list of strings.
        """
        self.input = InputHandler(input_data)
        self._reformat_data()

    def part_one(self) -> int:
        """Set part one data and return part one result.

        Returns
        -------
        result : int
        """
        self.__data = self._get_data_for_part_one()
        return self._part_one()

    def part_two(self) -> int:
        """Set part two data and return part two result.

        Returns
        -------
        result : int
        """
        self.__data = self._get_data_for_part_two()
        return self._part_two()

    @abstractmethod
    def _part_one(self) -> int:
        """Implement part one solution here"""

    @abstractmethod
    def _part_two(self) -> int:
        """Implement part two solution here"""

    def _reformat_data(self) -> None:
        """Change how the input data is formatted BEFORE any mutations."""

    def _get_data(self) -> list[Any]:
        """
        Process the input data to return a list.
        The return default is the input data as a list of strings, delimeted by newlines.
        This function is used to load the data for both part one and two if
        _get_data_for_part_one and _get_data_for_part_two are not implemented.

        You can change the implementation to return a list of whatever you want. For example::

            def _get_data(self) -> list[int]:
                return self.input.as_list(mutate=int)

        See how the solutions.input_reader.InputReader.as_list() method is used to
        process the import data.
        """
        return self.input.as_list()

    def _get_data_for_part_one(self) -> list[Any]:
        """
        Used if you need to have specific input processing for part one.
        See Solution._get_data to understand how to implement.
        """
        return self._get_data()

    def _get_data_for_part_two(self) -> list[Any]:
        """
        Used if you need to have specific input processing for part two.
        See Solution._get_data to understand how to implement.
        """
        return self._get_data()

    def __make_solution_result(
        self, func_name: str, timeit: bool, number: int
    ) -> SolutionResult:
        """Create a solution result.

        Parameters
        ----------
        func_name : str
            part_one or part_two

        timeit : bool
            Should the execution time be measured?

        number : int
            The number of repititions to measure the time.

        Returns
        -------
        SolutionResult
        """
        part_func = getattr(self, f"_{func_name}")
        data_func = getattr(self, f"_get_data_for_{func_name}")

        self.__data = data_func()
        return SolutionResult(
            result=part_func(),
            time=get_best_time(part_func, number=number) if timeit else None,
            number=number,
        )

    def __call__(
        self, timeit: bool = False, number: int = 1000
    ) -> tuple[SolutionResult, SolutionResult]:
        """Convenience function to get both part one and two results.

        Parameters
        ----------
        timeit : bool, optional
            Whether to time the execution of each part, by default False
        number : int, optional
            The number of repititions to time, by default 1000

        Returns
        -------
        tuple[SolutionResult, SolutionResult]
            Part one result, part two result
        """
        return (
            self.__make_solution_result("part_one", timeit, number),
            self.__make_solution_result("part_two", timeit, number),
        )

    def __str__(self) -> str:
        return f"Day {self.day}, {self.year}" + (f" - {self.name}" if self.name else "")



import os
from pathlib import Path
from typing import Optional

from aoc.helpers import read_asset


def read_solution_template() -> str:
    """Read the template file"""
    return read_asset("solution_template.txt")


def create_dir(dirpath: Path) -> None:
    """Create a directory"""
    if os.path.exists(dirpath):
        return

    os.mkdir(dirpath)


def save_file(filepath: Path, content: Optional[str] = None) -> None:
    """Save the content to disk"""
    if not filepath.parent.exists():
        create_dir(filepath.parent)

    with open(filepath, "w") as f:
        if content:
            f.write(content)


class SolutionManager:
    """Manage the creation of a solution"""

    year: int
    day: int
    name: Optional[str]

    def __init__(self, year: int, day: int, name: Optional[str] = None):
        self.year = year
        self.day = day
        self.name = name

    @property
    def solution_filepath(self) -> Path:
        filepath = os.path.join(f"{self.year}", f"day{self.day:02}.py")
        return Path(filepath)

    @property
    def has_solution_file(self) -> bool:
        """Check if the solution file exists"""
        return self.solution_filepath.exists()

    def create(self) -> "SolutionManager":
        """Create the solution file using the template"""

        if os.path.exists(self.solution_filepath):
            raise FileExistsError()

        template = self._create_solution_template()
        self.create_solution_file(template)
        return self

    def create_solution_file(self, template: str) -> None:
        """Save the solution file"""
        save_file(self.solution_filepath, template)

    def _create_solution_template(self) -> str:
        """Create a template for a solution and return the filepath"""
        template = read_solution_template()
        return self._format_solution_template(template)

    def _format_solution_template(self, template: str) -> str:
        """Format the template for a solution"""

        name = self.name if self.name else ""
        docstring = f"Day {self.day}" + (f": {self.name}" if self.name else "")
        return template.format(
            year=self.year, day=self.day, name=name, docstring=docstring
        )



from typing import Any
from typing import Callable
from typing import Optional


class InputHandler:
    """Class to reformat and mutate input data"""

    def __init__(self, input_data: list[str]):
        self.__content = [x.strip() for x in input_data]

    @property
    def content(self) -> list[str]:
        return self.__content

    def as_list(
        self,
        mutate: Optional[Callable[[str, Any], Any]] = None,
        **kwargs: Optional[Any]
    ) -> list[Any]:
        """Mutate each element in the input list

        Parameters
        ----------
        mutate : Optional[Callable], optional
            A function to apply to each element, by default None

        Returns
        -------
        list[Any]
            The mutated data
        """
        if not mutate:
            mutate = str.strip

        return [mutate(x, **kwargs) for x in self.content]  # type: ignore

    def reformat(
        self, formater: Callable[[list[str], Any], list[str]], **kwargs: Optional[Any]
    ) -> None:
        """Change the format of the input data

        You might want to do this if the original list of strings from the input data
        doesn't lend itself to the desired format. For example, if the input data is has groups
        of lines that should be together, but are seperated by newlines, you can reformat the data
        to make each group a single element in the list.

        See `aoc.reformaters`

        Parameters
        ----------
        formater : Callable
            A function that takes a list of strings and returns a new list of strings.
        """
        if self.__content:
            self.__content = formater(self.content, **kwargs)  # type: ignore



def combine_nonblank_lines(content: list[str], sep: str = " ") -> list[str]:
    content = [item.strip() for item in content]
    i, new_content, clean_content = 0, "", []

    while i < len(content):

        if content[i] == "":
            clean_content.append(new_content.strip())
            new_content = ""
        else:
            new_content += sep + content[i]

        i += 1

    if new_content:
        clean_content.append(new_content.strip())

    return clean_content



import os
import pathlib
from typing import Optional

import click

from aoc.helpers import display_solution
from aoc.helpers import get_latest_year
from aoc.helpers import instantiate_solution
from aoc.solution_manager import SolutionManager


@click.group()
def cli() -> None:
    """Advent of Code"""


@cli.command()
def init() -> None:
    """Initialize a new solution folder"""
    curdir = pathlib.Path(os.getcwd())

    if os.listdir(curdir):
        click.echo("Please run this command from an empty directory")
        return

    pathlib.Path("input.txt").touch()
    pathlib.Path("requirements.txt").touch()


@cli.command()
@click.argument("day", type=int)
@click.option("--year", type=int, default=get_latest_year())
@click.option("--timeit", is_flag=True, default=False)
@click.option("--number", "-n", "number", type=int, default=1000)
@click.option("-i", "input_file", type=str, default="input.txt")
@click.option("-s", "solution_file", type=str, default=None)
def run(
    day: int,
    year: int,
    timeit: bool,
    number: int,
    input_file: str,
    solution_file: Optional[str],
) -> None:
    """Run a solution file"""
    if not input_file:
        click.echo("Please specify an input file.")
        return

    if not solution_file:
        solution_file = f"{year}/day{day:02}.py"

    input_path = pathlib.Path(input_file)
    solution_path = pathlib.Path(solution_file)

    solution = instantiate_solution(solution_path, input_path)
    display_solution(solution, timeit, number)


@cli.command()
@click.argument("day", type=int)
@click.option("--year", type=int, default=get_latest_year())
@click.option("--name", type=str)
def create(year: int, day: int, name: Optional[str]) -> None:
    """Create a soution module"""
    try:
        manager = SolutionManager(year, day, name).create()
    except FileExistsError:
        click.echo(f"Solution for day {day}, {year} already exists!")
        return

    click.echo(f"Created solution template at {manager.solution_filepath}")
